# Handle.Net Performance Testing Software

The Handle.Net Testing software is a custom client software useful for evaluating the performance of the Handle.Net
Server software. Administration and resolution requests can be sent to the Handle.Net Server using this software. This
testing software can be simultaneously run from one or more machines for generating sufficient load on the Handle.Net
Server.

## High-Level Installation Procedure and Recommendations

1. Install the Handle.Net Server software v9.1.0 or later on its own machine. We recommend a memory of at least 500MB
   be allotted to the the server software process; in our tests, we allotted 4GB. Please refer to the Configure Test
   Server section below for details.
1. Install a cluster of test clients. In our tests, we started with a cluster of 10 clients (one client per machine) and
   varied that number to increase or decrease load on the server. Refer to the Create Test Client Cluster section below
   for details.
1. __DO NOT__ run this test client against live/production Handle.Net servers.
1. CNRI has conducted its own performance testing. Refer to https://handle.net/performance.html for
   discussion about CNRI's testing methodology, results, and recommendations for independent testing.

## Configure Test Server

To set up a test Handle.Net Server:

* Setup up a machine with Java 8 (or later) installed.
* Download and uncompress the standard handle server distribution (https://www.handle.net/download_hnr.html).
* Copy sample `server` directory from this software bundle to the test server. This will update the server storage with
  test handle records necessary for conducting the testing.
* Update the IP address and ports as necessary in `server/siteinfo.json`.
* To modify amount of memory allocated to the Handle.Net server, edit the startup script located at
  `./handle-9.1.0/bin/hdl`.
* Run the server with the following command:
    ./handle-9.1.0/bin/hdl-server server/
* The handle server should now be available and the various interfaces should resolve the test handles
  (12345/SMALL, 12345/LARGE, 12345/ADMIN).

## Create Test Client Cluster

In order to generate sufficient load to test the server, we recommend the test client be simultaneously run from a
cluster of machines. To increase the load on the test server, increase the number of client machines.

The test client is provided as a war file that can be deployed in any servlet container such as Apache Tomcat. The war
file can also be run in a hosted container service, such as Amazon Elastic Beanstalk to allow for easily scaling the
number of test clients. The test client issues administration and resolution requests to the Handle.Net Server. Details
of the number and nature of those requests (i.e., test configuration) can be sent to the test client via its own
interface.

For convenience, we have included configuration for setting up test clients on Amazon Web Services (AWS) Elastic
Beanstalk (EBS). To do so:

* Create an EBS app in the region of your choice on AWS.
* Create new Web Server Environment. Select Tomcat as the platform and upload the included war as the Application
  Code.
* Click the Configure More Options button.
* Click "Modify" on the Capacity box. Change the Environment type to "Load Balanced". Set the min and max instances
  to the same number. This ensures that the number of test clients does not change during the test. Click Save.
* Click "Modify" on the Software box. Add an Environment property called `ALLOWED_IPS`. This should be an IP
  address range that will be allowed to access your test clients on port 8080. For example, "10.0.0.0/24" would
  allow access from any machine with an IP address between 10.0.0.0 and 10.0.0.255, while "10.0.1.111/32" would only
  allow access from a machine with the IP address 10.0.1.111. Once this property is set, click Save. This should at
  least be the ip address of your local machine.
* Click Create Environment to start the environment. This process will take a few minutes.

Once the environment is created, tests can be run as described below. You can change the number of clients by changing
the Autoscaling configuration in the Capacity settings.

If the test client is deployed manually using a servlet container of your choice, make sure a sufficient number of
machines are used in order to fully load the server being tested.

The test client does not authenticate test request commands in any way, and will run tests based on the configuration
sent to it. For this reason, be sure to restrict access to your test clients. (`ALLOWED_IPS` configuration in EBS
settings is relevant.)

## Starting Tests

To initiate tests, send a HTTP POST request to the /test endpoint of each individual test client in the cluster. The
body of the request should be a JSON configuration for that test. The query parameter `startTime` can be used to
specify the time at which the clients should try to start sending requests. For example:

```
curl -XPOST -H "Content-Type: application/json" --data @testcreate.json http://TEST_CLIENT_IP:8000/test?startTime=1528394457000
```

Here `TEST_CLIENT_IP` is the IP address of a given test client. Refer to the EBS Specific Scripts section below to
leverage a few that we provide for running tests and compiling results when EBS is chosen to run the tests.

The test configuration JSON object supports the following fields:

* protocolString - The protocol for the test. One of UDP, TCP, HTTP, or API. Here, API is short for HTTP JSON API;
  whereas HTTP is short for native protocol tunneled through HTTP.
* requestMode - type of request. One of RESOLVE or CREATE.
* handle - The handle to resolve, or the base handle used for creates.
* authHandle - Handle of the admin user. The supplied admin handle is 12345/ADMIN.
* authIndex - Index of the admin user secret. The supplied admin private key is at index 300, and the password is
  at index 301.
* authKey - Private key of the admin user, in JSON Web Key format.
* authPassword - The admin user password. The supplied admin password is "apassword".
* numThreads - The number of threads the client should use for the test.
* iterationsPerThread - Number of iterations for each thread.
* delay - Delay between iterations.
* siteInfo - The siteinfo.json for the test server. This should be the actual JSON object, not a file path.

Sample config files are included in the `scripts` directory of this distribution. Note that the site info has to be
added to the samples prior to their use. Use the siteinfo.json from the server directory (possibly after changing the
IP address). The creation config file includes the appropriately formatted JWK private key.

Be sure not to run tests against live/production Handle.Net servers.

## Test Configuration Considerations

When sending a large number of request over TCP from a single machine, it is possible to exhaust the number of available
outgoing TCP ports. This cause performance to drop, as each request spends some time waiting for a port to become
available for use. There are a few variables that can be adjusted to avoid this issue:

* Number of client machines sending requests
* Number of threads per client machine (the `numThreads` setting)
* The delay between iterations on each thread (the `delay` setting)

The correct number for the threads and delay settings depends on the client machine specifications. We advise testing
to find out what works best for your given setup. In our experience, we found it best to set the number of threads
fairly low and have at least a 10ms delay. However, this requires increasing the number of client machines in order
to put sufficient load on the test server.

## Results Compilation

Each test client will send back a results JSON object with a following fields, as a response to the request initiated
to configure the test:

* startTime: Timestamp of when the test started
* totalRequestTimeInMillis: The sum of durations for each individual request
* requestCount: The number of (non-warmup) requests made.
* overallDurationInMillis: The time from test start to test stop.
* failureCount: Number of requests that failed.
* responses: An array of response object for each request, including start time, duration, and http response code.
* config: The config json that was sent to the server.

Using the response JSON objects from each test client, analysis may be performed to generate related performance metrics
of your choice.

## EBS Specific Scripts

Also included in the scripts directory are some sample scripts for run tests against multiple clients and processing
the results, when EBS environment is chosen for deploying the test clients. These scripts require Bash, Python (v2 or
v3), and the AWS command line tools installed and configured.

Once the Handle.Net server and test clients are deployed, tests can be run using the following command:

```
./do_test.sh <Name of Beanstalk Environment> <Beanstalk Env Region> <Config filename>
```

The results of the test will be saved into the results directory.

Run the script `process.py` to compile the results and produce an output table containing the following columns:

* Clients: The number of test client machine on which tests were run.
* Threads: The number of threads run on each client machine.
* Iters: The number of test iterations per thread.
* Delay: The delay that was inserted between each request.
* Proto: The protocol used for the test.
* Total ms: Sum of the runtime of all requests.
* Count: Number of requests made.
* Average ms: Average latency of the requests
* Reqs per sec: The overall throughput of the tests across all servers.
* 90% ms - Latency 90th percentile
* 95% ms - Latency 95th percentile
* 99% ms - Latency 99th percentile
