#!/usr/bin/env bash

ENV_NAME=$1
REGION=$2
CONFIGFILE=$3

SERVERS=$(aws ec2 describe-instances --query 'Reservations[*].Instances[*].[PublicIpAddress]' --filter Name=tag:Name,Values=${ENV_NAME} --output=text --region=${REGION})

FOLDER_NUM=1
while test -e "results/${FOLDER_NUM}"; do
    (( ++FOLDER_NUM ))
done
mkdir -p results/${FOLDER_NUM}

START_TIME=$(echo "($(date +%s) + 15) * 1000" | bc)
FOUND_TEST_SERVER=0
CURL_PIDS=""
for SERVER in $SERVERS; do
    if [[ $SERVER != "None" ]]; then
        FOUND_TEST_SERVER=1
        echo "Starting test on: ${SERVER}"
        curl -XPOST -H "Content-Type: application/json" --data @${CONFIGFILE} http://${SERVER}:8080/test?startTime=${START_TIME} 2> /dev/null > results/${FOLDER_NUM}/${SERVER}.out &
        CURL_PIDS="${CURL_PIDS} ${!}"
    fi
done

if [[ $FOUND_TEST_SERVER == 0 ]]; then
    echo "No test servers found for environment \"${ENV_NAME}\" in region \"${REGION}\""
    exit 1
fi

if [[ ! $CURL_PIDS == "" ]]; then
    echo -n "Waiting for results."
    while kill -0 $CURL_PIDS 2>/dev/null; do
        sleep 2
        echo -n "."
    done
    echo ""
fi

echo "Requests complete. Output folder: ./results/${FOLDER_NUM}"
python process.py results/${FOLDER_NUM}
#say "test complete"
